
__name__ = 'tinybird-cli'
__description__ = 'Tinybird Command Line Tool'
__url__ = 'https://docs.tinybird.co/cli.html'
__author__ = 'Tinybird'
__author_email__ = 'support@tinybird.co'
__version__ = '1.0.0b66'
__revision__ = '6c1e4b29c'
